"""Contains the basic / standard Todo class definition."""

# pylint: disable=format-string-without-interpolation

from __future__ import annotations

import abc
import datetime as dt
from functools import total_ordering
import re
from typing import Any, Dict, Final, Generic, List, Optional, Tuple, cast
import uuid

from eris import ErisError, Err, Ok, Result
from metaman import cname

from ._common import (
    CONTEXT_PREFIX,
    DEFAULT_PRIORITY,
    EPIC_PREFIX,
    PROJECT_PREFIX,
    PUNCTUATION,
    RE_DATE,
    from_date,
    is_metadata_tag,
    is_prefix_tag,
    to_date,
)
from .types import Metadata, Priority, T


RE_TODO: Final = r"""
(?P<x>x[ ]+)?                        # optional 'x'
(?:\((?P<priority>[A-Z])\)[ ]+)?     # priority
(?:
    (?:(?P<done_date>{0})[ ]+)?      # optional date of completion
    (?:(?P<create_date>{0})[ ]+)     # optional date of creation
)?
(?P<desc>[A-Za-z0-9+@].*)            # description
""".format(
    RE_DATE
)


class TodoMixin(Generic[T], abc.ABC):
    """Implements standard Todo-like behaviors.."""

    @property
    def ident(self) -> str:
        """Unique identifier."""
        key = "_uuid_ident"
        if not hasattr(self, key):
            setattr(self, key, uuid.uuid4())

        result: str = getattr(self, key)
        return result

    def __repr__(self: T) -> str:  # noqa: D105
        kwargs: Dict[str, Any] = {}

        if self.contexts:
            kwargs["contexts"] = self.contexts

        if self.create_date is not None:
            kwargs["create_date"] = self.create_date

        if self.done_date is not None:
            kwargs["done_date"] = self.done_date

        if self.done:
            kwargs["done"] = self.done

        if self.epics:
            kwargs["epics"] = self.epics

        if self.metadata:
            kwargs["metadata"] = self.metadata

        if self.priority != DEFAULT_PRIORITY:
            kwargs["priority"] = repr(self.priority)

        if self.projects:
            kwargs["projects"] = self.projects

        if kwargs:
            pretty_kwargs = ", " + ", ".join(
                f"{k}={v}" for (k, v) in kwargs.items()
            )
        else:
            pretty_kwargs = ""

        return f"{cname(self)}(desc={self.desc!r}{pretty_kwargs})"

    def __hash__(self: T) -> int:  # noqa: D105
        return hash(self.metadata.values())

    def __eq__(self: T, other: object) -> bool:  # noqa: D105
        if not isinstance(other, type(self)):  # pragma: no cover
            return False

        return all(
            [
                self.desc == other.desc,
                self.contexts == other.contexts,
                self.create_date == other.create_date,
                self.done_date == other.done_date,
                self.done == other.done,
                self.metadata == other.metadata,
                self.priority == other.priority,
                self.projects == other.projects,
            ]
        )

    def __lt__(self: T, other: object) -> bool:  # noqa: D105
        if not isinstance(other, type(self)):
            raise ValueError(
                f"Unable to compare '{type(other)}' object with 'Todo' object."
            )

        if self.done != other.done:
            return not self.done and other.done

        if self.priority != other.priority:
            return self.priority < other.priority

        if self.done_date is None and other.done_date is not None:
            return True

        if other.done_date is None and self.done_date is not None:
            return False

        if self.done_date is not None and other.done_date is not None:
            if self.done_date != other.done_date:
                return self.done_date < other.done_date
            elif (self_dtime := self.metadata.get("dtime", None)) and (
                other_dtime := other.metadata.get("dtime", None)
            ):
                assert isinstance(self_dtime, str)
                assert isinstance(other_dtime, str)
                return self_dtime < other_dtime

        if self.create_date and other.create_date:
            if self.create_date == other.create_date:
                if (
                    (self_ctime := self.metadata.get("ctime", None))
                    and (other_ctime := other.metadata.get("ctime", None))
                    and self_ctime != other_ctime
                ):
                    assert isinstance(self_ctime, str)
                    assert isinstance(other_ctime, str)
                    return self_ctime < other_ctime
            else:
                return self.create_date < other.create_date

        if (self_id := self.metadata.get("id", None)) and (
            other_id := other.metadata.get("id", None)
        ):
            assert isinstance(self_id, str)
            assert isinstance(other_id, str)
            return self_id < other_id

        return self.desc < other.desc


@total_ordering
class Todo(TodoMixin):
    """Represents a single task in a todo list."""

    def __init__(
        self,
        desc: str,
        *,
        contexts: Tuple[str, ...] = (),
        create_date: dt.date = None,
        done_date: dt.date = None,
        done: bool = False,
        epics: Tuple[str, ...] = (),
        metadata: Metadata = None,
        priority: Priority = DEFAULT_PRIORITY,
        projects: Tuple[str, ...] = (),
    ):
        self.contexts = contexts
        self.create_date = create_date
        self.desc = desc
        self.done_date = done_date
        self.done = done
        self.epics = epics
        self.metadata = metadata or {}
        self.priority = priority
        self.projects = projects

    @classmethod
    def from_line(cls, line: str) -> Result[Todo, ErisError]:
        """Contructs a Todo object from a string (usually a line in a file).

        Args:
            line: The line to use to construct our new Todo object.
        """
        line = line.strip()

        re_todo_match = re.match(RE_TODO, line, re.VERBOSE)
        if re_todo_match is None:
            return Err(
                f"The provided string ({line!r}) does not appear to properly"
                " adhere to the todo.txt format. See"
                " https://github.com/todotxt/todo.txt for the specification."
            )

        done: bool = False
        if re_todo_match.group("x"):
            done = True

        priority: Priority = DEFAULT_PRIORITY
        if grp := re_todo_match.group("priority"):
            priority = cast(Priority, grp)

        create_date: Optional[dt.date] = None
        if grp := re_todo_match.group("create_date"):
            create_date = to_date(grp)

        done_date: Optional[dt.date] = None
        if grp := re_todo_match.group("done_date"):
            done_date = to_date(grp)

        desc = re_todo_match.group("desc")
        all_words = desc.split(" ")

        project_list: List[str] = []
        context_list: List[str] = []
        epics_list: List[str] = []
        for some_list, prefix in [
            (project_list, PROJECT_PREFIX),
            (context_list, CONTEXT_PREFIX),
            (epics_list, EPIC_PREFIX),
        ]:
            for word in all_words:
                if is_prefix_tag(prefix, word):
                    value = word[len(prefix) :]
                    value = _clean_value(value)
                    if value not in some_list:
                        some_list.append(value)

        projects = tuple(project_list)
        contexts = tuple(context_list)
        epics = tuple(epics_list)

        metadata: Metadata = {}
        for word in all_words:
            if is_metadata_tag(word):
                kv = word.split(":", maxsplit=1)
                key, value = kv
                value = _clean_value(value)

                if key in metadata:
                    continue

                metadata[key] = value

        todo = cls(
            contexts=contexts,
            create_date=create_date,
            desc=desc,
            done_date=done_date,
            done=done,
            epics=epics,
            metadata=metadata,
            priority=priority,
            projects=projects,
        )
        return Ok(todo)

    def to_line(self) -> str:
        """Converts this Todo object back to a line."""
        result = ""
        if self.done:
            result += "x "

        if self.priority != DEFAULT_PRIORITY:
            result += f"({self.priority}) "

        if self.done_date is not None:
            result += from_date(self.done_date) + " "

        if self.create_date is not None:
            result += from_date(self.create_date) + " "

        result += self.desc

        return result

    def new(self, **kwargs: Any) -> Todo:
        """Creates a new Todo using the current Todo's attrs as defaults."""
        contexts = kwargs.get("contexts", self.contexts)
        create_date = kwargs.get("create_date", self.create_date)
        desc = kwargs.get("desc", self.desc)
        done_date = kwargs.get("done_date", self.done_date)
        done = kwargs.get("done", self.done)
        epics = kwargs.get("epics", self.epics)
        metadata = kwargs.get("metadata", self.metadata)
        priority: Priority = kwargs.get("priority", self.priority)
        projects = kwargs.get("projects", self.projects)
        return Todo(
            contexts=contexts,
            create_date=create_date,
            desc=desc,
            done_date=done_date,
            done=done,
            epics=epics,
            metadata=metadata,
            priority=priority,
            projects=projects,
        )


def _clean_value(word: str) -> str:
    """Cleanup context, metadata, or project value.

    Makes the following changes to `word`:

      - Strips any punctuation from the right-side of `word`.
      - Removes any possesive apostrophe at the end of `word`.

    NOTE: Will not strip punctuation if `word` is composed ONLY of punctuation
      characters.
    """
    result = word.rstrip(PUNCTUATION)
    if not result:
        result = word

    result = result.split("'", maxsplit=1)[0]
    return result



"""Tests for the magodo package."""

from __future__ import annotations

from typing import Sequence, cast

from pytest import mark

from magodo import DEFAULT_PRIORITY, Todo
from magodo._common import to_date
from magodo.types import AbstractTodo

from .shared import MOCK_TODO_KWARGS, MagicTodo, assert_todos_equal


params = mark.parametrize


@params(
    "line,expected_todo",
    [
        ("no priority todo", Todo(desc="no priority todo")),
        (f"({DEFAULT_PRIORITY}) basic todo", Todo(desc="basic todo")),
        (
            f"({DEFAULT_PRIORITY}) 2022-01-10 todo with create date",
            Todo(
                desc="todo with create date",
                create_date=to_date("2022-01-10"),
            ),
        ),
        (
            f"({DEFAULT_PRIORITY}) 2022-03-04 2022-01-10 todo with done date",
            Todo(
                desc="todo with done date",
                create_date=to_date("2022-01-10"),
                done_date=to_date("2022-03-04"),
            ),
        ),
        (
            f"x ({DEFAULT_PRIORITY}) 2022-01-10 done todo",
            Todo(
                desc="done todo",
                create_date=to_date("2022-01-10"),
                done=True,
            ),
        ),
        (
            "x (A) todo with priority",
            Todo(
                desc="todo with priority",
                done=True,
                priority="A",
            ),
        ),
        (
            f"({DEFAULT_PRIORITY}) todo for +some +project",
            Todo(
                desc="todo for +some +project",
                projects=("some", "project"),
            ),
        ),
        (
            f"({DEFAULT_PRIORITY}) todo for +some +project and a @context.",
            Todo(
                desc="todo for +some +project and a @context.",
                projects=("some", "project"),
                contexts=("context",),
            ),
        ),
        (
            f"({DEFAULT_PRIORITY}) todo with +some meta:data and a @context"
            " due:2022-12-31",
            Todo(
                desc="todo with +some meta:data and a @context due:2022-12-31",
                projects=("some",),
                contexts=("context",),
                metadata={"meta": "data", "due": "2022-12-31"},
            ),
        ),
        (
            f"({DEFAULT_PRIORITY}) todo with some dep:123,10,20,30, another"
            " dep:456, and a 3rd dep:789... foo:bar @crazy",
            Todo(
                desc=(
                    "todo with some dep:123,10,20,30, another dep:456, and a"
                    " 3rd dep:789... foo:bar @crazy"
                ),
                contexts=("crazy",),
                metadata={"dep": "123,10,20,30", "foo": "bar"},
            ),
        ),
        (
            "x:1030 2022-01-12 Some done todo with a @ctx.",
            Todo(
                desc="x:1030 2022-01-12 Some done todo with a @ctx.",
                contexts=("ctx",),
                metadata={"x": "1030"},
            ),
        ),
        (
            "o +regression @test Do +duplicate +duplicate @keys @keys foo:FOO"
            " foo:bar foo:baz foo:boom bar:BAR still show when they should be"
            " ignored?",
            MagicTodo(
                Todo(
                    desc=(
                        "Do duplicate duplicate keys keys still show when"
                        " they should be ignored? | @keys @test +duplicate"
                        " +regression bar:BAR foo:FOO"
                    ),
                    contexts=(
                        "test",
                        "keys",
                    ),
                    projects=(
                        "regression",
                        "duplicate",
                    ),
                    metadata={"foo": "FOO", "bar": "BAR"},
                )
            ),
        ),
    ],
)
def test_todo(line: str, expected_todo: AbstractTodo) -> None:
    """Test the Todo type."""
    actual = cast(MagicTodo, type(expected_todo).from_line(line).unwrap())

    actual = actual.new(**MOCK_TODO_KWARGS)
    expected_todo = expected_todo.new(**MOCK_TODO_KWARGS)

    assert_todos_equal(actual, expected_todo)


@params(
    "todos,idxs",
    [
        (
            [
                Todo("baz", priority="M"),
                Todo("foo", priority="G"),
                Todo("o bar"),
            ],
            [1, 0, 2],
        ),
        (
            [Todo("o foo"), Todo("o bar"), Todo("o baz", done=True)],
            [1, 0, 2],
        ),
        (
            [
                Todo("o foo"),
                Todo("o foo"),
                Todo("o A", done_date=to_date("2022-01-01")),
                Todo(
                    "o B",
                    done_date=to_date("1999-01-01"),
                    metadata={"dtime": "0123"},
                ),
                Todo(
                    "o C",
                    done_date=to_date("2022-01-01"),
                    metadata={"dtime": "1100"},
                ),
                Todo(
                    "o D",
                    done_date=to_date("2022-01-01"),
                    metadata={"dtime": "1000"},
                ),
            ],
            [0, 1, 3, 2, 5, 4],
        ),
    ],
)
def test_sort(todos: Sequence[Todo], idxs: Sequence[int]) -> None:
    """Test that Todo objects sort properly."""
    assert len(todos) == len(idxs)
    N = len(todos)

    expected = [todos[idxs[i]] for i in range(N)]
    assert sorted(todos) == expected

    magic_todos = [MagicTodo(todo) for todo in todos]
    magic_expected = [magic_todos[idxs[i]] for i in range(N)]
    assert sorted(magic_todos) == magic_expected



"""Test that TodoGroup objects work as they should."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from pytest import mark, param
from syrupy.assertion import SnapshotAssertion as Snapshot

from magodo import TodoGroup

from .data import get_all_todo_paths
from .shared import MOCK_TODO_KWARGS, MagicTodo


params = mark.parametrize


@params("todo_file", get_all_todo_paths())
@params(
    "filter_kwargs",
    [
        param({}, id="no-filter-kwargs"),
        param({"desc": "Double colons"}, id="filter-kwarg-desc"),
        param({"contexts": ["high"]}, id="ctx"),
    ],
)
def test_group_from_path(
    snapshot: Snapshot, todo_file: Path, filter_kwargs: dict[str, Any]
) -> None:
    """Test the TodoGroup.from_path() function."""
    todo_group = TodoGroup.from_path(MagicTodo, todo_file)
    todo_group = todo_group.filter_by(**filter_kwargs)
    assert (
        sorted(repr(todo.new(**MOCK_TODO_KWARGS)) for todo in todo_group)
        == snapshot
    )
